
import argparse
import datetime
import numpy as np
import time
import torch
import torch.backends.cudnn as cudnn
import json
import os
from pathlib import Path
from timm.data.mixup import Mixup
from timm.models import create_model
from timm.utils import ModelEma
from CL_framework.CL_datasets import create_image_downstream_datasets, create_fewshot_dataset
from beit3_tools.optim_factory import create_optimizer, get_parameter_groups, \
    LayerDecayValueAssigner, get_is_head_flag_for_vit
from CL_framework import modelin
from CL_framework.engine import train_one_epoch, get_handler, evaluate, fewshot_evaluate
from beit3_tools.beit3_datasets import create_downstream_dataset
from beit3_tools.utils import NativeScalerWithGradNormCount as NativeScaler
from beit3_tools import utils
import wandb
import datasets
import warnings
warnings.filterwarnings("ignore")


def get_args():
    parser = argparse.ArgumentParser('BEiT fine-tuning and evaluation script for image classification', add_help=False)

    # Model parameters
    parser.add_argument('--model', default='beit_base_patch16_224', type=str, metavar='MODEL',
                        help='Name of model to train')
    parser.add_argument('--task', type=str, required=True,
                        choices=['cifar10', 'cifar100','imagenet', 'bird',
                                 'flowers102', 'oxfordpet', 'caltech256',
                                 'miniImageNet', 'tieredImageNet', 'CIFAR-FS', 'FC100',
                                 'places365', 'inat2021', 'inat2017',],
                        help='Name of task to fine-tuning')

    parser.add_argument('--input_size', default=224, type=int,
                        help='images input size')
    parser.add_argument('--drop_path', type=float, default=0.1, metavar='PCT',
                        help='Drop path rate (default: 0.1)')

    parser.add_argument('--checkpoint_activations', action='store_true', default=None,
                        help='Enable checkpointing to save your memory.')
    parser.add_argument('--sentencepiece_model', type=str, required=True,
                        help='Sentencepiece model path for the pretrained model.')
    parser.add_argument('--vocab_size', type=int, default=64010)
    parser.add_argument('--num_max_bpe_tokens', type=int, default=64)

    parser.add_argument('--model_ema', action='store_true', default=False)
    parser.add_argument('--model_ema_decay', type=float, default=0.9999, help='')
    parser.add_argument('--model_ema_force_cpu', action='store_true', default=False, help='')

    # Optimizer parameters
    parser.add_argument('--opt', default='adamw', type=str, metavar='OPTIMIZER',
                        help='Optimizer (default: "adamw"')
    parser.add_argument('--opt_eps', default=1e-8, type=float, metavar='EPSILON',
                        help='Optimizer Epsilon (default: 1e-8)')
    parser.add_argument('--opt_betas', default=[0.9, 0.999], type=float, nargs='+', metavar='BETA',
                        help='Optimizer Betas (default: 0.9, 0.999, use opt default)')
    parser.add_argument('--clip_grad', type=float, default=None, metavar='NORM',
                        help='Clip gradient norm (default: None, no clipping)')
    parser.add_argument('--momentum', type=float, default=0.9, metavar='M',
                        help='SGD momentum (default: 0.9)')
    parser.add_argument('--weight_decay', type=float, default=0.05,
                        help='weight decay (default: 0.05)')

    parser.add_argument('--lr', type=float, default=5e-4, metavar='LR',
                        help='learning rate (default: 5e-4)')
    parser.add_argument('--layer_decay', type=float, default=0.9)
    parser.add_argument('--task_head_lr_weight', type=float, default=0)

    parser.add_argument('--warmup_lr', type=float, default=1e-6, metavar='LR',
                        help='warmup learning rate (default: 1e-6)')
    parser.add_argument('--min_lr', type=float, default=1e-6, metavar='LR',
                        help='lower lr bound for cyclic schedulers that hit 0 (1e-6)')
    parser.add_argument('--warmup_epochs', type=int, default=5, metavar='N',
                        help='epochs to warmup LR, if scheduler supports')
    parser.add_argument('--warmup_steps', type=int, default=-1, metavar='N',
                        help='num of steps to warmup LR, will overload warmup_epochs if set > 0')

    parser.add_argument('--batch_size', default=64, type=int)
    parser.add_argument('--eval_batch_size', default=None, type=int)
    parser.add_argument('--epochs', default=20, type=int)
    parser.add_argument('--update_freq', default=1, type=int)
    parser.add_argument('--save_ckpt_freq', default=5, type=int)

    # Augmentation parameters
    parser.add_argument('--randaug', action='store_true', default=False)
    parser.add_argument('--train_interpolation', type=str, default='bicubic',
                        help='Training interpolation (random, bilinear, bicubic default: "bicubic")')

    # Finetuning params
    parser.add_argument('--finetune', default='',
                        help='finetune from checkpoint')
    parser.add_argument('--model_key', default='model|module', type=str)
    parser.add_argument('--model_prefix', default='', type=str)

    # Dataset parameters
    parser.add_argument('--data_path', default='/datasets01/imagenet_full_size/061417/', type=str,
                        help='dataset path')

    parser.add_argument('--output_dir', default='',
                        help='path where to save, empty for no saving')
    parser.add_argument('--log_dir', default=None,
                        help='path where to tensorboard log')
    parser.add_argument('--device', default='cuda',
                        help='device to use for training / testing')
    parser.add_argument('--seed', default=0, type=int)
    parser.add_argument('--resume', default='',
                        help='resume from checkpoint')
    parser.add_argument('--auto_resume', action='store_true')
    parser.add_argument('--no_auto_resume', action='store_false', dest='auto_resume')
    parser.set_defaults(auto_resume=True)

    parser.add_argument('--save_ckpt', action='store_true')
    parser.add_argument('--no_save_ckpt', action='store_false', dest='save_ckpt')
    parser.set_defaults(save_ckpt=True)

    parser.add_argument('--start_epoch', default=0, type=int, metavar='N',
                        help='start epoch')
    parser.add_argument('--eval', action='store_true',
                        help='Perform evaluation only')
    parser.add_argument('--dist_eval', action='store_true', default=False,
                        help='Enabling distributed evaluation')
    parser.add_argument('--num_workers', default=10, type=int)
    parser.add_argument('--pin_mem', action='store_true',
                        help='Pin CPU memory in DataLoader for more efficient (sometimes) transfer to GPU.')
    parser.add_argument('--no_pin_mem', action='store_false', dest='pin_mem')
    parser.set_defaults(pin_mem=True)

    # distributed training parameters
    parser.add_argument('--world_size', default=1, type=int,
                        help='number of distributed processes')
    parser.add_argument('--local_rank', default=-1, type=int)
    parser.add_argument('--dist_on_itp', action='store_true')
    parser.add_argument('--dist_url', default='env://',
                        help='url used to set up distributed training')

    # parameter for dump predictions (VQA, COCO captioning, NoCaps)
    parser.add_argument('--task_cache_path', default=None, type=str)

    # parameter for imagenet finetuning
    parser.add_argument('--nb_classes', default=1000, type=int,
                        help='number of the classification types')
    parser.add_argument('--mixup', type=float, default=0,
                        help='mixup alpha, mixup enabled if > 0.')
    parser.add_argument('--cutmix', type=float, default=0,
                        help='cutmix alpha, cutmix enabled if > 0.')
    parser.add_argument('--cutmix_minmax', type=float, nargs='+', default=None,
                        help='cutmix min/max ratio, overrides alpha and enables cutmix if set (default: None)')
    parser.add_argument('--mixup_prob', type=float, default=1.0,
                        help='Probability of performing mixup or cutmix when either/both is enabled')
    parser.add_argument('--mixup_switch_prob', type=float, default=0.5,
                        help='Probability of switching to cutmix when both mixup and cutmix enabled')
    parser.add_argument('--mixup_mode', type=str, default='batch',
                        help='How to apply mixup/cutmix params. Per "batch", "pair", or "elem"')

    # augmentation parameters for imagenet finetuning
    parser.add_argument('--color_jitter', type=float, default=0.4, metavar='PCT',
                        help='Color jitter factor (default: 0.4)')
    parser.add_argument('--aa', type=str, default='rand-m9-mstd0.5-inc1', metavar='NAME',
                        help='Use AutoAugment policy. "v0" or "original". " + "(default: rand-m9-mstd0.5-inc1)')
    parser.add_argument('--smoothing', type=float, default=0.1,
                        help='Label smoothing (default: 0.1)')

    # evaluation parameters for imagenet
    parser.add_argument('--crop_pct', type=float, default=None)

    # random Erase params for imagenet finetuning
    parser.add_argument('--reprob', type=float, default=0.25, metavar='PCT',
                        help='Random erase prob (default: 0.25)')
    parser.add_argument('--remode', type=str, default='pixel',
                        help='Random erase mode (default: "pixel")')
    parser.add_argument('--recount', type=int, default=1,
                        help='Random erase count (default: 1)')
    parser.add_argument('--resplit', action='store_true', default=False,
                        help='Do not random erase first (clean) augmentation split')

    # parameter for captioning finetuning
    parser.add_argument('--captioning_mask_prob', type=float, default=0.6)
    parser.add_argument('--drop_worst_ratio', type=float, default=0.2)
    parser.add_argument('--drop_worst_after', type=int, default=12000)
    parser.add_argument('--num_beams', type=int, default=3)
    parser.add_argument('--length_penalty', type=float, default=0.6)

    # label smoothing for imagenet and captioning
    parser.add_argument('--label_smoothing', type=float, default=0.1)

    # deepspeed parameters
    parser.add_argument('--enable_deepspeed', action='store_true', default=False)
    parser.add_argument('--initial_scale_power', type=int, default=16)
    parser.add_argument('--zero_stage', default=0, type=int,
                        help='ZeRO optimizer stage (default: 0)')

    # distributed parameters

    # wandb options
    parser.add_argument('--enable_wandb', action='store_true', default=False, )

    # dataloading options
    parser.add_argument('--load_image_from_huggingface_hub', action='store_true', default=False, )
    parser.add_argument('--cluster', action='store_true', default=False, )

    # image classfication
    parser.add_argument('--n_cls', default=None, type=int,
                        help='Number of classes')
    parser.add_argument('--train_ratio', type=float, default=0.9,
                        help='length of training set')
    parser.add_argument('--eval_freq', default=10, type=int,
                        help='Number of classes')
    # Cl Vision options
    parser.add_argument('--temp', type=float, default=0.07,
                        help='temperature for loss function')
    parser.add_argument('--alpha', type=float, default=0.5,
                        help='ratio of cl loss')
    parser.add_argument('--loss', type=str, default='CE',
                        choices=['SupCon', 'SimCLR', 'CE', 'SCHaNe'], help='choose method')
    parser.add_argument('--tau_plus', default=0.1, type=float, help='Positive class priorx')
    parser.add_argument('--k', default=200, type=int, help='Top k most similar images used to predict the label')
    parser.add_argument('--estimator', default='hard', type=str, help='Choose loss function')
    parser.add_argument('--beta', default=1.0, type=float, help='Choose loss function')

    # few-shot learning
    parser.add_argument('--eval_fewshot', action='store_true', default=False, )
    parser.add_argument('--n_test_runs', type=int, default=3000, metavar='N',
                        help='Number of test runs')
    parser.add_argument('--n_ways', type=int, default=5, metavar='N',
                        help='Number of classes for doing each classification run')
    parser.add_argument('--n_shots', type=int, default=1, metavar='N',
                        help='Number of shots in test')
    parser.add_argument('--n_queries', type=int, default=15, metavar='N',
                        help='Number of query in test')
    parser.add_argument('--n_aug_support_samples', default=5, type=int,
                        help='The number of augmented samples for each meta test sample')
    parser.add_argument('--use_trainval', action='store_true', help='use trainval set')

    known_args, _ = parser.parse_known_args()

    if known_args.enable_deepspeed:
        try:
            import deepspeed
            from deepspeed import DeepSpeedConfig
            parser = deepspeed.add_config_arguments(parser)
            ds_init = deepspeed.initialize
        except:
            print("Please 'pip install deepspeed==0.4.0'")
            exit(0)
        raise Exception("Windows does not support DeepSpeed")
    else:
        ds_init = None

    return parser.parse_args(), ds_init


def main(args, ds_init):
    utils.init_distributed_mode(args)

    if ds_init is not None:
        utils.create_ds_config(args)

    if args.task_cache_path is None:
        args.task_cache_path = args.output_dir

    print(args)


    device = torch.device(args.device)

    # fix the seed for reproducibility
    seed = args.seed + utils.get_rank()
    torch.manual_seed(seed)
    np.random.seed(seed)
    # random.seed(seed)

    cudnn.benchmark = True

    # datasets.config.IN_MEMORY_MAX_SIZE = 16000000000
    print('Set dataset loading memory size to:', datasets.config.IN_MEMORY_MAX_SIZE)
    #  set output dir
    if args.eval:
        exp_tag = f"eval_{args.task}_seed{args.seed}_{args.n_shots}nshot_{args.alpha}alpha_{args.finetune.split('/')[-1]}/"
    else:
        exp_tag = f"{args.model}_{args.loss}loss_seed{args.seed}_{args.alpha}alpha_{args.temp}temp_{args.epochs}epochs_{args.n_shots}nshot_{args.task}tasks_{args.finetune.split('/')[-1]}/"

    args.log_dir = os.path.join(args.log_dir, exp_tag)

    if utils.get_rank() == 0 and args.log_dir is not None:
        os.makedirs(args.log_dir, exist_ok=True)
        log_writer = utils.TensorboardLogger(log_dir=args.log_dir)
    else:
        log_writer = None

    if args.n_cls is None:
        if args.task == 'cifar10':
            args.n_cls = 10
        elif args.task == 'cifar100':
            args.n_cls = 100
        elif args.task == 'imagenet':
            args.n_cls = 1000
        elif args.task == 'bird':
            args.n_cls = 200
        elif args.task == 'caltech256':
            args.n_cls = 257
        elif args.task == 'flowers102':
            args.n_cls = 102
        elif args.task == 'oxfordpet':
            args.n_cls = 37
        elif args.task == 'miniImageNet' and args.use_trainval:
            args.n_cls = 80
        elif args.task == 'miniImageNet' and not args.use_trainval:
            args.n_cls = 64
        elif args.task == 'tieredImageNet' and args.use_trainval:
            args.n_cls = 448
        elif args.task == 'tieredImageNet' and not args.use_trainval:
            args.n_cls = 351
        elif args.task in ['CIFAR-FS', 'FC100'] and args.use_trainval:
            args.n_cls = 80
        elif args.task == 'CIFAR-FS' and not args.use_trainval:
            args.n_cls = 64
        elif args.task == 'FC100' and not args.use_trainval:
            args.n_cls = 60
        elif args.task == 'places365':
            args.n_cls = 365
        elif args.task == 'inat2021':
            args.n_cls = 10000
        elif args.task == 'inat2017':
            args.n_cls = 5089
        elif args.task == 'inat2018':
            args.n_cls = 8142
        else:
            raise ValueError('dataset not supported: {}'.format(args.task))

    if args.task in ['miniImageNet', 'tieredImageNet','CIFAR-FS', 'FC100']:
        args.data_aug = True
        data_loader_train, data_loader_val = create_fewshot_dataset(args)

    else:
        data_loader_train, data_loader_val = create_image_downstream_datasets(args)


    args.output_dir = os.path.join(args.output_dir, exp_tag)

    if utils.is_main_process():
        os.makedirs(args.output_dir, exist_ok=True)

    if not args.model.endswith(args.task):
        if args.task in ['cifar100','imagenet', 'path', 'bird','flowers102', 'cifar10',
                         'oxfordpet', 'caltech256', 'miniImageNet', 'tieredImageNet',
                         'CIFAR-FS', 'FC100', 'places365', 'inat2021', 'inat2017', 'inat2018']:
            model_config = "%s_imageclassification" % args.model
        else:
            raise NotImplementedError
    else:
        model_config = args.model

    # initial wandb
    if utils.is_main_process() and args.enable_wandb:
        wandb_config = {
            'lr': args.lr,
            'min_lr': args.min_lr,
            'weight_decay': args.weight_decay,
            'batch_size': args.batch_size,
            'epochs': args.epochs,
            'warmup_epochs': args.warmup_epochs,
            'layer_decay': args.layer_decay,
            'fine_tune': args.finetune,
        }
        wandb.init(project='beit3_classification',
                   config=wandb_config, save_code=True)

    # print("model_config = %s" % model_config)
    model = create_model(
        model_config,
        pretrained=False,
        drop_path_rate=args.drop_path,
        vocab_size=args.vocab_size,
        checkpoint_activations=args.checkpoint_activations,
        n_cls=args.n_cls,
        temp=args.temp,
        loss=args.loss,
    )

    if args.finetune:
        utils.load_model_and_may_interpolate(args.finetune, model, args.model_key, args.model_prefix)

    model.to(device)

    model_ema = None
    if args.model_ema:
        # Important to create EMA model after cuda(), DP wrapper, and AMP but before SyncBN and DDP wrapper
        model_ema = ModelEma(
            model,
            decay=args.model_ema_decay,
            device='cpu' if args.model_ema_force_cpu else '',
            resume='')
        print("Using EMA with decay = %.8f" % args.model_ema_decay)

    model_without_ddp = model
    n_parameters = sum(p.numel() for p in model.parameters() if p.requires_grad)

    # print("Model = %s" % str(model_without_ddp))
    print('number of params:', n_parameters)


    total_batch_size = args.batch_size * args.update_freq * utils.get_world_size()
    num_training_steps_per_epoch = len(data_loader_train.dataset) // total_batch_size
    print("LR = %.8f" % args.lr)
    print("Batch size = %d" % total_batch_size)
    print("Update frequent = %d" % args.update_freq)
    print("Number of training examples = %d" % len(data_loader_train.dataset))
    print("Number of training training per epoch = %d" % num_training_steps_per_epoch)


    num_layers = model_without_ddp.get_num_layers()
    if args.layer_decay < 1.0:
        lrs = list(args.layer_decay ** (num_layers + 1 - i) for i in range(num_layers + 2))
        assigner = LayerDecayValueAssigner(lrs)
    elif args.task_head_lr_weight > 1:
        assigner = LayerDecayValueAssigner([1.0, args.task_head_lr_weight], scale_handler=get_is_head_flag_for_vit)
    else:
        assigner = None

    if assigner is not None:
        print("Assigned values = %s" % str(assigner.values))


    skip_weight_decay_list = model.no_weight_decay()


    if args.distributed:
        torch.distributed.barrier()
    if args.enable_deepspeed:
        loss_scaler = None
        optimizer_params = get_parameter_groups(
            model, args.weight_decay, skip_weight_decay_list,
            assigner.get_layer_id if assigner is not None else None,
            assigner.get_scale if assigner is not None else None)
        model, optimizer, _, _ = ds_init(
            args=args, model=model, model_parameters=optimizer_params,
            dist_init_required=not args.distributed,
        )

        print("model.gradient_accumulation_steps() = %d" % model.gradient_accumulation_steps())
        assert model.gradient_accumulation_steps() == args.update_freq
    else:
        if args.distributed:
            print("Using torch.nn.parallel.DistributedDataParallel ... on GPU %d" % args.gpu)
            model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[args.gpu], find_unused_parameters=True)
            model_without_ddp = model.module

        optimizer = create_optimizer(
            args, model_without_ddp, skip_list=skip_weight_decay_list,
            get_num_layer=assigner.get_layer_id if assigner is not None else None,
            get_layer_scale=assigner.get_scale if assigner is not None else None)
        loss_scaler = NativeScaler()


    lr_schedule_values = utils.cosine_scheduler(
        args.lr, args.min_lr, args.epochs, num_training_steps_per_epoch,
        warmup_epochs=args.warmup_epochs, warmup_steps=args.warmup_steps,
    )

    utils.auto_load_model(
        args=args, model=model, model_without_ddp=model_without_ddp,
        optimizer=optimizer, loss_scaler=loss_scaler, model_ema=model_ema)

    task_handler = get_handler(args)

    # mixup for imagenet
    mixup_fn = None

    if args.eval and args.eval_fewshot:
        meta_testloader, meta_valloader = create_fewshot_dataset(args, eval=True)
        model.to(device)
        if args.distributed:
            torch.distributed.barrier()
            model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[args.gpu], find_unused_parameters=True)
            model_without_ddp = model.module
        start = time.time()
        test_acc_feat, test_std_feat, \
            = fewshot_evaluate(data_loader=meta_testloader, model=model,
                               device=device, handler=task_handler, )
        test_time = time.time() - start
        print('test_acc_feat: {:.4f}, val_std: {:.4f}, time: {:.1f}'.format(test_acc_feat,
                                                                            test_std_feat,
                                                                            test_time))

        start = time.time()
        val_acc_feat, test_std_feat \
            = fewshot_evaluate(data_loader=meta_valloader, model=model,
                               device=device, handler=task_handler, )
        val_time = time.time() - start
        print('val_acc_feat: {:.4f}, val_std: {:.4f}, time: {:.1f}'.format(val_acc_feat,
                                                                           test_std_feat,
                                                                           val_time))
        # store result into json file
        if args.distributed:
            torch.distributed.barrier()
        if utils.is_main_process():
            json_file = f"{args.output_dir}/{args.task}_fewshot_test_results.json"
            with open(json_file, mode="a+", encoding="utf-8") as writer:
                writer.write(json.dumps({"test_acc_feat": test_acc_feat,
                                         "test_std_feat": test_std_feat,
                                         "test_time": test_time,
                                         "val_acc_feat": val_acc_feat,
                                         "val_std_feat": test_std_feat,
                                         "val_time": val_time,
                                         }, indent=None))
                writer.write('\n')

        exit(0)


    print(f"Start training for {args.epochs} epochs")
    start_time = time.time()

    if args.eval_fewshot:
        meta_testloader, meta_valloader = create_fewshot_dataset(args, eval=True)

    max_accuracy = 0.0
    best_test_acc = 0.0
    test_std = 0.0
    best_val_acc = 0.0
    val_std = 0.0
    for epoch in range(args.start_epoch, args.epochs):
        if args.distributed:
            data_loader_train.sampler.set_epoch(epoch)
        if log_writer is not None:
            log_writer.set_step(epoch * num_training_steps_per_epoch * args.update_freq)
        train_stats = train_one_epoch(
            model, data_loader_train, optimizer, device, task_handler, epoch,
            epoch * num_training_steps_per_epoch, lr_schedule_values, loss_scaler,
            args.clip_grad, args.update_freq, model_ema, log_writer, args.task, mixup_fn,
        )
        if args.output_dir and args.save_ckpt:
            if (epoch + 1) % args.save_ckpt_freq == 0 or epoch + 1 == args.epochs:
                utils.save_model(
                    args=args, model=model, model_without_ddp=model_without_ddp, optimizer=optimizer,
                    loss_scaler=loss_scaler, epoch=epoch, model_ema=model_ema)
        if data_loader_val is not None and (epoch + 1) % args.eval_freq == 0 and not args.eval_fewshot:
            if args.task not in ["coco_captioning", "nocaps"]:
                test_stats, task_key = evaluate(data_loader_val, model, device, task_handler)

            print(
                f"Performance of the network on the {len(data_loader_val.dataset)} val images: {test_stats[task_key]:.1f}%")
            if max_accuracy < test_stats[task_key]:
                max_accuracy = test_stats[task_key]
                if args.output_dir and args.save_ckpt:
                    utils.save_model(
                        args=args, model=model, model_without_ddp=model_without_ddp, optimizer=optimizer,
                        loss_scaler=loss_scaler, epoch="best", model_ema=model_ema)

            print(f'Max performance: {max_accuracy:.2f}%')
            if log_writer is not None:
                log_writer.update(acc=test_stats[task_key], head="perf", step=epoch)

            log_stats = {**{f'train_{k}': v for k, v in train_stats.items()},
                         **{f'val_{k}': v for k, v in test_stats.items()},
                         'epoch': epoch,
                         'n_parameters': n_parameters}
        else:
            log_stats = {**{f'train_{k}': v for k, v in train_stats.items()},
                         # **{f'test_{k}': v for k, v in test_stats.items()},
                         'epoch': epoch,
                         'n_parameters': n_parameters}

        if utils.is_main_process() and args.enable_wandb:
            wandb.log(log_stats)

        if args.output_dir and utils.is_main_process():
            if log_writer is not None:
                log_writer.flush()
            with open(os.path.join(args.output_dir, "log.txt"), mode="a+", encoding="utf-8") as f:
                f.write(json.dumps(log_stats) + "\n")

        if args.eval_fewshot and (epoch + 1) % args.eval_freq == 0:


            start = time.time()
            test_acc_feat, test_std_feat, \
                = fewshot_evaluate(data_loader=meta_testloader, model=model,
                                   device=device, handler=task_handler, )
            test_time = time.time() - start
            print('test_acc_feat: {:.4f}, val_std: {:.4f}, time: {:.1f}'.format(test_acc_feat,
                                                                                test_std_feat,
                                                                                test_time))
            if test_acc_feat > best_test_acc:
                best_test_acc = test_acc_feat
                test_std = test_std_feat
                print('best_test_acc Updataed: {:.4f}, test_std: {:.4f}'.format(best_test_acc, test_std))

            start = time.time()
            val_acc_feat, test_std_feat \
                = fewshot_evaluate(data_loader=meta_valloader, model=model,
                                   device=device, handler=task_handler, )
            val_time = time.time() - start
            print('val_acc_feat: {:.4f}, val_std: {:.4f}, time: {:.1f}'.format(val_acc_feat,
                                                                               test_std_feat,
                                                                               val_time))
            if val_acc_feat > best_val_acc:
                best_val_acc = val_acc_feat
                val_std = test_std_feat

    total_time = time.time() - start_time
    total_time_str = str(datetime.timedelta(seconds=int(total_time)))
    print('Training time {}'.format(total_time_str))

    if args.eval_fewshot:
        # print best acc
        print('best_test_acc: {:.4f}, test_std: {:.4f}'.format(best_test_acc, test_std))
        print('best_val_acc: {:.4f}, val_std: {:.4f}'.format(best_val_acc, val_std))




    if utils.is_main_process() and args.enable_wandb:
        wandb.finish()


if __name__ == '__main__':
    opts, ds_init = get_args()
    if opts.output_dir:
        Path(opts.output_dir).mkdir(parents=True, exist_ok=True)
    main(opts, ds_init)
